%{

#include <stdio.h>
#include <stdlib.h>

#include "memory.h"
#include "instruction.h"
#include "cpu.h"
#include "label.h"

#define	NOT_DEFINED_LABEL_MARK	(MEMORY_SIZE<<1)
#define	to_undefined_label( L ) (L+NOT_DEFINED_LABEL_MARK)
#define	un_undefined_label( L ) (L-NOT_DEFINED_LABEL_MARK)
#define	is_undefined_label( L ) (L>=NOT_DEFINED_LABEL_MARK)

static int memory[MEMORY_SIZE];
static int pc = 0;

extern int yylex();
extern int yyparse();
extern FILE* yyin;

extern char *yytext;
extern int yylineno;

void yyerror(const char* s);
%}

%union {
  int ival;			/* 整数値 */
  char *sval;		/* 文字列 */
}

%token<ival> T_INT
%token<sval> T_ID
%token T_COMMA
%token T_COLON
%token T_NEWLINE

%token T_NOP
%token T_HALT
%token T_OUT
%token T_IN
%token T_JUMP
%token T_JZ
%token T_JN

%token T_LD
%token T_ST

%token T_ADD
%token T_SUB

%token T_CALL
%token T_RET

%token T_INC
%token T_DEC

%token	T_A
%token	T_IX

%token	T_LBRA
%token	T_RBRA

%type<ival> constant address memory

%start program

%%

program: commands
	;

commands:
	|	commands command T_NEWLINE
	;

command:
	|	label_def
	|	instruction
	|	label_def instruction
	|	error
		{
		  	yyclearin;
			yyerrok;
		}
	;

label_def	:	T_ID T_COLON
	{
	  if ( search_label ( $1 ) == LABEL_NOT_FOUND ) {
		set_label ( $1, pc );
	  } else {
		int label = get_label ( $1 );

		if ( !is_undefined_label ( label ) ) {
		  fprintf ( stderr, "label(%s) already defined\n", $1 );
		} else {
		  int addr = un_undefined_label ( label );

		  while ( addr != NOT_DEFINED_LABEL_MARK ) {
			int next = memory[addr+1];

			memory[addr+1] = pc;
			addr = next;
		  }
		  set_label ( $1, pc );
		}
	  }
	}
	;

constant:	T_INT
		{ $$ = $1; }
	;

address:	constant
		{ $$ = $1; }
	|		T_ID
		{
		  /*
			定義済
				定義値を使えばよい
			未定義
				初めて
				二度目以後
		   */
			if ( search_label ( $1 ) == LABEL_NOT_FOUND ) {
			  /* 未定義で初めて */
			  $$ = NOT_DEFINED_LABEL_MARK;
			  set_label ( $1, to_undefined_label ( pc ) );
			} else {
				int addr = get_label ( $1 );

				if ( is_undefined_label ( addr ) ) {
				  $$ = un_undefined_label ( addr );
				  set_label ( $1, to_undefined_label ( pc ) );
				  /* label は、最後に未定義で参照された場所 */
				} else {
				  /* 定義済 */
				  $$ = addr;
				}
			}
		}
	;

memory:	T_LBRA address T_RBRA
		{ $$ = $2; }
	;

instruction
	:	T_NOP
	{
	  memory[pc++] = CPU_NOP;
	}
	|	T_HALT
	{
	  memory[pc++] = CPU_HALT;
	}
	|	T_OUT	T_A
	{
	  memory[pc++] = CPU_OUT;
	}
	|	T_IN	T_A
	{
	  memory[pc++] = CPU_IN;
	}
	|	T_JUMP address
	{
	  memory[pc++] = CPU_JUMP;
	  memory[pc++] = $2;
	}
	|	T_JZ  address
	{
	  memory[pc++] = CPU_JZ;
	  memory[pc++] = $2;
	}
	|	T_JN  address
	{
	  memory[pc++] = CPU_JN;
	  memory[pc++] = $2;
	}
	|	T_LD T_A T_COMMA memory
	{
	  memory[pc++] = CPU_LDM;
	  memory[pc++] = $4;
	}
	|	T_ST memory T_COMMA T_A
	{
	  memory[pc++] = CPU_STM;
	  memory[pc++] = $2;
	}
	|	T_LD T_A T_COMMA constant
	{
	  memory[pc++] = CPU_LDI;
	  memory[pc++] = $4;
	}
	|	T_ADD T_A T_COMMA memory
	{
	  memory[pc++] = CPU_ADD;
	  memory[pc++] = $4;
	}
	|	T_SUB T_A T_COMMA memory
	{
	  memory[pc++] = CPU_SUB;
	  memory[pc++] = $4;
	}
	|	T_CALL address
	{
	  memory[pc++] = CPU_CALL;
	  memory[pc++] = $2;
	}
	|	T_RET
	{
	  memory[pc++] = CPU_RET;
	}
	;
%%

int main(int argc, char *argv[]) {
	int i;
	CPU	cpu;
	int debug = 0;

	if ( argc > 1 && !strcmp ( argv[1], "-d" ) ) {
	  debug = 1;
	  argc--;
	  argv++;
	}

	if ( argc == 2 ) {
	  if ( ( yyin = fopen ( argv[1], "r" ) ) == NULL ) {
		fprintf ( stderr, "file(%s) not found\n", argv[1] );
		exit ( -1 );
	  }
	} else {
		yyin = stdin;
	}

	do { 
		yyparse();
	} while(!feof(yyin));

	if ( argc == 2 ) {
	  fclose ( yyin );
	}

	init_cpu ( &cpu, 0 );

	execute ( &cpu, memory, debug );

	return 0;
}

void yyerror(const char* s) {

	fprintf ( stderr,
		"%d: (%s) Parse error: %s\n",
		yylineno, yytext, s );
	//	exit(1);
}
